/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wSolver.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTri3.h"
#include "wTimer.h"
#include "wTag.h"
#include "wFct2.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wSource.h"
#include "wBoundary.h"
#include "wDirichlet.h"
#include "wDisplayHook.h"
#include "wPeriodic.h"
#include "wHeatTerm.h"
#include "wResults.h"
#include "wMshExport.h"
#include "wLinearSolver.h"

#include <Eigen/Sparse>
#include <typeinfo>
#include <algorithm>
#include <iomanip>
#include <deque>

#include <tbb/global_control.h>
#include <tbb/parallel_for_each.h>
#include <tbb/spin_mutex.h>
using namespace heat;
using namespace fwk;

#undef min
#undef max

Solver::Solver(std::shared_ptr<Problem> _pbl, std::shared_ptr<tbox::LinearSolver> _linsol) : pbl(_pbl), linsol(_linsol)
{
    dhook = NULL;

    // default values
    nthreads = 1;
    verbose = 1;
    stopit = false;
    restol = 1e-2;
    itmax = 100;
    save = true;

    // update element memory
    pbl->initElems();
}

Solver::~Solver()
{
    std::cout << "~Solver()\n";
}

/**
 * @todo planter clairement si pas de Medium
 * @todo calculer la taille du systeme en fct du nombre de noeuds des physical groups utilisés
 */
void Solver::start(MshExport *mshWriter)
{
    tbb::spin_mutex mutex;
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    // start timers
    tms["total"].start();
    tms["prepro"].start(); // ----- start preprocessing phase

    if (verbose > 1)
        std::cout << *this;

    auto msh = pbl->msh;

    // setup results
    T1.resize(msh->nodes.size(), 0.); // in case of restart
    Eigen::Map<Eigen::VectorXd> T1_(T1.data(), T1.size());

    Results results;
    results.scalars_at_nodes["T"] = &T1;

    // initial temperature guess
    if (!pbl->dBCs.empty())
    {
        double tmin = std::numeric_limits<double>::max();
        double tmax = std::numeric_limits<double>::min();
        for (auto dBC : pbl->dBCs)
        {
            double v = dBC->val;
            if (v < tmin)
                tmin = v;
            if (v > tmax)
                tmax = v;
        }
        /*
        std::cout << "tmin=" << tmin << "\n";
        std::cout << "tmax=" << tmax << "\n";
        */
        for (auto n : msh->nodes)
            T1[n->row] = (tmin + tmax) / 2;

        for (auto dBC : pbl->dBCs)
            dBC->apply(T1);
    }

    // periodic BCs
    if (pbl->pdic)
    {
        if (!pbl->dBCs.empty())
            throw std::runtime_error("cannot manage periodic BCs and Dirichlet BCs simultaneously!\n");
        pbl->pdic->init();

        // initial guess for T
        for (auto n : msh->nodes)
            T1[n->row] = pbl->pdic->TM;
    }
    tms["prepro"].stop(); // ----- end of preprocessing phase

    // --------------------------------------------------------------------
    // Iteration LOOP
    // --------------------------------------------------------------------

    int it = 0;
    while (it <= itmax)
    {
        std::vector<double> rhs(msh->nodes.size());                                             // dummy RHS
        Eigen::Map<Eigen::VectorXd> rhs_ = Eigen::Map<Eigen::VectorXd>(rhs.data(), rhs.size()); // actual RHS

        tms["Kass"].start();
        // K matrix assembly
        Eigen::SparseMatrix<double, Eigen::RowMajor> K(msh->nodes.size(), msh->nodes.size());
        buildK(K, rhs_);
        // tomatlab("K.txt", K);
        // throw std::runtime_error("stop!");
        tms["Kass"].stop();

        tms["BCs"].start();

        if (!pbl->pdic)
        {
            builds(rhs_); // sources
            buildq(rhs_); // prescribed fluxes (=sources over lines)

            // BCs
            for (auto dBC : pbl->dBCs) // apply Dirichlet BCs to RHS
                dBC->apply(rhs);
        }
        tms["BCs"].stop();
        // std::cout << rhs << "\n";

        // residual calculation
        tms["residual"].start();

        // method #1
        Eigen::VectorXd qint = K * T1_; // calcul de q_int = K T1:

        // method #2
        if (0)
        {
            Eigen::VectorXd qint2;
            buildqint(qint2); // calcul de q_int par integration du flux sur l'elem
            double normdiff = (qint - qint2).norm();
            // if(verbose>1)
            std::cout << "norm diff=" << normdiff << '\n';
        }

        /*
        for(auto i=0; i<msh->nodes.size(); ++i)
            std::cout << i << ":" << qint[i] << "," << qint2[i] << "\n";
        */

        Eigen::VectorXd res = qint - rhs_; // residual
        double rnorm2 = res.norm();
        double rnorminf = res.lpNorm<Eigen::Infinity>();
        tms["residual"].stop();

        if (verbose > 0)
        {
            std::cout << std::fixed << std::setprecision(3);
            std::cout << " * it #" << std::setw(3) << std::left << it
                      << " |r|2 = " << std::setw(9) << std::scientific << std::right << rnorm2
                      << "  |r|inf = " << std::setw(9) << std::scientific << std::right << rnorminf << '\n';
        }
        // check convergence
        if (rnorm2 < restol && it != 0)
            break;
        it++;

        // solve
        tms["solve"].start();
        if (verbose > 1)
            std::cout << "solving...\n";
        if (it == 1)
        {
            linsol->analyze(K);
        }
        linsol->factorize(K);
        linsol->solve(rhs_, T1_);
        tms["solve"].stop();
    }
    if (it > itmax)
    {
        std::cout << "\n**WARNING! iterative process failed to converge\n";
        std::cout << "           in less than " << itmax << " iterations!\n\n"; // TODO: generer une erreur!
    }

    // --------------------------------------------------------------------
    // Postprocessing
    // --------------------------------------------------------------------

    if (verbose > 1)
        std::cout << "postpro...\n";
    tms["post"].start();

    // setup results

    std::vector<double> gradT_x(msh->elems.size());         ///< gradT_x
    std::vector<double> gradT_y(msh->elems.size());         ///< gradT_y
    std::vector<Eigen::Vector3d> gradT(msh->elems.size());  ///< gradT
    std::vector<Eigen::Vector3d> kgradT(msh->elems.size()); ///< gradT
    std::vector<double> kmoy11(msh->elems.size());          ///< mean k11 / element
    std::vector<double> kmoy22(msh->elems.size());          ///< mean k22 / element
    std::vector<double> kmoy12(msh->elems.size());          ///< mean k12 / element
    std::vector<double> kmoy21(msh->elems.size());          ///< mean k21 / element
    std::vector<Eigen::MatrixXd> kmoy(msh->elems.size());
    for (auto &km : kmoy)
        km = Eigen::Matrix3d::Zero();

    results.scalars_at_elems["gradT_x"] = &gradT_x;
    results.scalars_at_elems["gradT_y"] = &gradT_y;
    results.scalars_at_elems["k11"] = &kmoy11;
    results.scalars_at_elems["k22"] = &kmoy22;
    results.scalars_at_elems["k12"] = &kmoy12;
    results.scalars_at_elems["k21"] = &kmoy21;
    results.vectors_at_elems["gradT"] = &gradT;
    results.vectors_at_elems["q [heat flux]"] = &kgradT;
    results.tensors_at_elems["k [tensor]"] = &kmoy;

    double Vtot = 0.0;
    Eigen::Vector2d qM = Eigen::Vector2d::Zero();
    for (auto mat : pbl->media)
    {
        tbb::parallel_for_each(
            mat->tag->elems.begin(), mat->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::TRI3)
                    return;
                // std::cout << "processing element #" << e->no << "\n";

                // (flux moyen . volume) sur l'element
                Eigen::VectorXd qV = HeatTerm::computeFlux(*e, T1, *mat->k);

                double V = e->getVol();

                if (save)
                {
                    // gradient of T at GP 0
                    Eigen::VectorXd grad = e->computeGradient(T1, 0);
                    int i = e->no - 1;
                    gradT_x[i] = grad(0);
                    gradT_y[i] = grad(1);
                    gradT[i] = Eigen::Vector3d(grad(0), grad(1), 0);

                    // mean q over the element
                    kgradT[i] = Eigen::Vector3d(qV(0) / V, qV(1) / V, 0);

                    // mean k over the element
                    Eigen::MatrixXd kmean = HeatTerm::computeMatrix(*e, T1, *mat->k) / V;
                    kmoy11[i] = kmean(0, 0);
                    kmoy22[i] = kmean(1, 1);
                    kmoy12[i] = kmean(0, 1);
                    kmoy21[i] = kmean(1, 0);

                    // tensor
                    kmean.conservativeResize(3, 3);
                    kmean.row(2) = Eigen::RowVector3d::Zero();
                    kmean.col(2) = Eigen::Vector3d::Zero();
                    kmoy[i] = kmean;
                    // std::cout << kmean << '\n';
                }

                tbb::spin_mutex::scoped_lock lock(mutex);
                Vtot += V;
                qM += qV; // qM = qM + qV
            });
    }
    qM /= Vtot;

    if (verbose > 1)
    {
        std::cout << "Vtot = " << Vtot << '\n';
        std::cout << "qM = " << qM << '\n';
    }

    // periodic BCs
    if (pbl->pdic)
    {
        pbl->pdic->qM = Eigen::Vector3d(qM(0), qM(1), 0.);
        if (verbose > 1)
        {
            std::cout << "=> k_x = " << ((pbl->pdic->dTxM == 0.0) ? 0 : qM(0) / pbl->pdic->dTxM) << '\n';
            std::cout << "=> k_y = " << ((pbl->pdic->dTyM == 0.0) ? 0 : qM(1) / pbl->pdic->dTyM) << '\n';
        }
    }

    if (save)
    {
        if (verbose > 1)
            std::cout << "saving results to disk...\n";
        mshWriter->save(msh->name, results);
    }

    tms["post"].stop();
    tms["total"].stop();

    if (verbose > 0)
    {
        std::cout << "---CPU statistics ---\n";
        std::cout << tms;
        std::cout << "---------------------\n";
    }
}

void Solver::buildK(Eigen::SparseMatrix<double, Eigen::RowMajor> &K2, Eigen::Map<Eigen::VectorXd> &rhs)
{
    tbb::spin_mutex mutex;

    std::shared_ptr<MshData> msh = pbl->msh;

    // List of triplets to build matrix
    std::deque<Eigen::Triplet<double>> T;

    // periodic BCs
    // (on va sommer les lignes des noeuds "d" sur celles des noeuds "n")

    std::vector<int> rows(msh->nodes.size());
    for (size_t i = 0; i < msh->nodes.size(); ++i)
        rows[i] = msh->nodes[i]->row;

    if (pbl->pdic)
    {
        for (auto np : pbl->pdic->BT)
            rows[np.d->row] = np.n->row;
        for (auto np : pbl->pdic->LR)
            rows[np.d->row] = np.n->row;
    }

    if (verbose > 1)
        std::cout << "building K (TBB/lambda) using " << nthreads << " threads...\n";

    bool useTBB = false;
    if (useTBB)
    {
        // 1-STEP PROCEDURE

        for (auto mat : pbl->media)
        {
            if (verbose > 1)
                std::cout << "\tprocessing " << *mat << '\n';
            tbb::parallel_for_each(
                mat->tag->elems.begin(), mat->tag->elems.end(),
                [&](Element *e)
                {
                    if (e->type() != ElType::TRI3)
                        return;
                    // std::cout << "processing element #" << e->no << "\n";

                    Eigen::MatrixXd Ke = HeatTerm::build(*e, T1, *mat->k, false);

                    // assembly
                    tbb::spin_mutex::scoped_lock lock(mutex);
                    for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                    {
                        Node *nodi = e->nodes[ii];
                        for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                        {
                            Node *nodj = e->nodes[jj];
                            T.push_back(Eigen::Triplet<double>(rows[nodi->row], nodj->row, Ke(ii, jj)));
                        }
                    }
                });
        }
    }
    else
    {
        // 3-STEP PROCEDURE

        // *** build Job list - "fake run"

        if (verbose > 1)
            std::cout << "building job list...\n";

        for (auto mat : pbl->media)
        {
            if (verbose > 1)
                std::cout << "\tprocessing " << *mat << '\n';
            std::for_each(mat->tag->elems.begin(), mat->tag->elems.end(),
                          [&](Element *e)
                          {
                              if (e->type() != ElType::TRI3)
                                  return;
                              Eigen::MatrixXd Ke = HeatTerm::build(*e, T1, *mat->k, true); // bidon
                          });
        }

        // *** execute all jobs generated during the fake run

        if (verbose > 1)
            std::cout << "executing job list...\n";
        for (auto mat : pbl->media)
            mat->k->evalall();

        // ***  perform the assembly with the results ("true run")

        if (verbose > 1)
            std::cout << "performing assembly...\n";

        for (auto mat : pbl->media)
        {
            if (verbose > 1)
                std::cout << "\tprocessing " << *mat << '\n';

            std::for_each(mat->tag->elems.begin(), mat->tag->elems.end(), [&](Element *e) { // devrait pouvoir etre fait en TBB...
                if (e->type() != ElType::TRI3)
                    return;
                Eigen::MatrixXd Ke = HeatTerm::build(*e, T1, *mat->k, false);

                // assembly
                // tbb::spin_mutex::scoped_lock lock(mutex);
                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        T.push_back(Eigen::Triplet<double>(rows[nodi->row], nodj->row, Ke(ii, jj)));
                    }
                }
            });
        }
        if (verbose > 1)
            std::cout << "assembly done.\n";
    }
    // Build matrix without BCs
    K2.setFromTriplets(T.begin(), T.end());

    // periodic BCs
    if (pbl->pdic)
    {
        // equations \theta_d (R,T) fct (L,B)
        double lx = pbl->pdic->corners[1]->pos(0) - pbl->pdic->corners[0]->pos(0);
        double ly = pbl->pdic->corners[3]->pos(1) - pbl->pdic->corners[0]->pos(1);
        if (verbose > 1)
            std::cout << "lx = " << lx << '\n';
        if (verbose > 1)
            std::cout << "ly = " << ly << '\n';
        double TM = pbl->pdic->TM;
        double dTxM = pbl->pdic->dTxM;
        double dTyM = pbl->pdic->dTyM;

        for (auto np : pbl->pdic->LR)
        {
            // clears related row
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K2, np.d->row); it; ++it)
                it.valueRef() = 0.;

            K2.coeffRef(np.d->row, np.d->row) = 1.0;
            K2.coeffRef(np.d->row, np.n->row) = -1.0;
            rhs(np.d->row) = lx * dTxM;
        }

        for (auto np : pbl->pdic->BT)
        {
            // clears related row
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K2, np.d->row); it; ++it)
                it.valueRef() = 0.;

            K2.coeffRef(np.d->row, np.d->row) = 1.0;
            K2.coeffRef(np.d->row, np.n->row) = -1.0;
            rhs(np.d->row) = ly * dTyM;
        }

        // corners \theta_2,3,4 fct \theta_1

        for (auto i = 0; i < 4; ++i)
        {
            Node *nod = pbl->pdic->corners[i];
            // clears related row
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K2, nod->row); it; ++it)
                it.valueRef() = 0.;
        }

        Node *n1 = pbl->pdic->corners[0];
        for (auto i = 1; i < 4; ++i)
        {
            Node *ni = pbl->pdic->corners[i];
            K2.coeffRef(ni->row, ni->row) = 1.0;
            K2.coeffRef(ni->row, n1->row) = -1.0;
        }
        Node *n2 = pbl->pdic->corners[1];
        Node *n3 = pbl->pdic->corners[2];
        Node *n4 = pbl->pdic->corners[3];
        rhs(n2->row) = lx * dTxM;
        rhs(n3->row) = lx * dTxM + ly * dTyM;
        rhs(n4->row) = ly * dTyM;

        // thermal energy consistency condition (n1->row)

        double Vtot = 0.0;
        double rhocM_Vtot = 0;
        for (auto mat : pbl->media)
        {
            // std::cout << "\tprocessing " << *mat << '\n';
            tbb::parallel_for_each(
                mat->tag->elems.begin(), mat->tag->elems.end(),
                [&](Element *e)
                {
                    if (e->type() != ElType::TRI3)
                        return;
                    // std::cout << "processing element #" << e->no << "\n";

                    // ** Ce matrix => K matrix
                    Eigen::VectorXd Ce = HeatTerm::build(*e, T1, Fct0C(1.0));

                    // assembly
                    tbb::spin_mutex::scoped_lock lock(mutex);
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        K2.coeffRef(n1->row, nodj->row) += (mat->rhoc) * Ce(jj);

                        Vtot += Ce(jj);
                        rhocM_Vtot += (mat->rhoc) * Ce(jj);
                    }
                });
        }
        rhs(n1->row) = rhocM_Vtot * TM;
        if (verbose > 1)
            std::cout << "Vtot=" << Vtot << "\n";
        if (verbose > 1)
            std::cout << "rhocM_Vtot=" << rhocM_Vtot << "\n";
    }

    if (!pbl->dBCs.empty())
    {
        if (verbose > 1)
            std::cout << "applying Dirichlet BCs...\n";

        // apply Dirichlet BCs to K
        for (auto dBC : pbl->dBCs)
        {
            for (auto nod : dBC->nodes)
            {
                // std::cout << "processing " << *nod << "\n";
                for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K2, nod->row); it; ++it)
                {
                    if (it.row() == it.col())
                        it.valueRef() = 1.;
                    else
                        it.valueRef() = 0.;
                }
            }
        }
        if (verbose > 1)
            std::cout << "done." << std::endl;
    }

    // Clean matrix and turn to compressed row format
    K2.prune(0.);
    K2.makeCompressed();

    if (verbose > 1)
    {
        std::cout << "K (" << K2.rows() << "," << K2.cols() << ") nnz=" << K2.nonZeros() << "\n";
    }
}

void Solver::buildqint(Eigen::VectorXd &qint)
{
    tbb::spin_mutex mutex;

    qint = Eigen::VectorXd::Zero(pbl->msh->nodes.size());

    for (auto mat : pbl->media)
    {
        if (verbose > 1)
            std::cout << "\tprocessing " << *mat << '\n';
        tbb::parallel_for_each(
            mat->tag->elems.begin(), mat->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::TRI3)
                    return;
                // std::cout << "processing element #" << e->no << "\n";

                Eigen::VectorXd qe = HeatTerm::build2(*e, T1, *mat->k);

                // assembly
                tbb::spin_mutex::scoped_lock lock(mutex);
                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    qint(nodi->row) += qe(ii);
                }
            });
    }

    if (!pbl->dBCs.empty())
    {
        if (verbose > 1)
            std::cout << "applying Dirichlet BCs...\n";

        // apply Dirichlet BCs to qint
        for (auto dBC : pbl->dBCs)
        {
            for (auto nod : dBC->nodes)
            {
                qint(nod->row) = T1[nod->row]; // pour comparer avec K.T1
            }
        }
        if (verbose > 1)
            std::cout << "done." << std::endl;
    }
}

/**
 * @brief volume sources
 */

void Solver::builds(Eigen::Map<Eigen::VectorXd> &s)
{
    tbb::spin_mutex mutex;

    if (verbose > 1)
        std::cout << "building s (TBB/lambda) using " << nthreads << " threads...\n";

    for (auto src : pbl->srcs)
    {
        // std::cout << "\tprocessing " << *src << '\n';
        tbb::parallel_for_each(
            src->tag->elems.begin(), src->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::TRI3 && e->type() != ElType::LINE2)
                    return;
                // std::cout << "processing element #" << e->no << "\n";

                // ** se vector => s vector
                Eigen::VectorXd se = HeatTerm::build(*e, T1, *src->f);

                // assembly
                tbb::spin_mutex::scoped_lock lock(mutex);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    s(nodi->row) += se(ii);
                }
            });
    }

    if (verbose > 1)
    {
        std::cout << "s (" << s.size() << ")\n";
    }
}

/**
 * @brief terme BC Neumann (pourrait etre mergé avec "builds"... c'est idem)
 * @todo merge this routine with builds
 */
void Solver::buildq(Eigen::Map<Eigen::VectorXd> &q)
{
    tbb::spin_mutex mutex;

    if (verbose > 1)
        std::cout << "building q (TBB/lambda) using " << nthreads << " threads...\n";

    for (auto bnd : pbl->bnds)
    {
        // std::cout << "\tprocessing " << *src << '\n';
        tbb::parallel_for_each(
            bnd->tag->elems.begin(), bnd->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::LINE2)
                    return;
                // std::cout << "processing element #" << e->no << "\n";

                // ** qe vector => q vector
                Eigen::VectorXd qe = HeatTerm::build(*e, T1, *bnd->f);

                // assembly
                tbb::spin_mutex::scoped_lock lock(mutex);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    q(nodi->row) += qe(ii);
                }
            });
    }

    if (verbose > 1)
    {
        std::cout << "q (" << q.size() << ")\n";
    }
}

void Solver::write(std::ostream &out) const
{
    out << "heat::Solver"
        << "\n";
}
